import tempfile
import unittest
from pathlib import Path

from ..distinfo_inference import analyze, iterparents


class DistinfoInferenceTest(unittest.TestCase):
    def test_iterparents(self) -> None:
        self.assertEqual(["bad.wolf", "bad"], list(iterparents("bad.wolf.foo")))

    def test_analyze_basic(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)
            (pd / "RECORD").write_text(
                """\
foo/bar.py,
foo/__init__.py,
../x.py,
"""
            )
            dist = analyze(pd)
            self.assertEqual({"foo", "foo.bar"}, dist.provided_names)
            self.assertEqual(set(), dist.namespace_names)
            self.assertEqual({"foo"}, dist.minimal_names)

    def test_analyze_new_namespace_packages(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)
            (pd / "RECORD").write_text(
                """\
foo/bar/baz.py,
foo/__init__.py,
"""
            )
            dist = analyze(pd)
            self.assertEqual({"foo", "foo.bar.baz"}, dist.provided_names)
            self.assertEqual({"foo.bar"}, dist.namespace_names)
            # TODO: not sure I'm 100% confident in this; it needs a docstring to
            # better define what "minimal names" means when there's a
            # non-toplevel namespace.
            self.assertEqual({"foo.bar.baz"}, dist.minimal_names)

    def test_analyze_setuptools_namespace(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)
            (pd / "RECORD").write_text(
                """\
foo/bar/baz.py,
foo/bar/__init__.py,
foo/__init__.py,
"""
            )
            (pd / "namespace_packages.txt").write_text(
                """\
foo
foo.bar



"""
            )  # some extra newlines on purpose
            dist = analyze(pd)
            self.assertEqual({"foo.bar.baz"}, dist.provided_names)
            self.assertEqual({"foo", "foo.bar"}, dist.namespace_names)
            self.assertEqual({"foo.bar.baz"}, dist.minimal_names)



import unittest

from ..distinfo import iter_all_distinfo_dirs
from ..distinfo_inference import analyze


class IterDistinfoDirsTest(unittest.TestCase):
    def test_basic(self) -> None:
        for project, version, subdir in iter_all_distinfo_dirs():
            if project == "click":
                break
        else:
            self.fail("Could not find click distinfo")

        dist = analyze(subdir)
        self.assertEqual({"click"}, dist.minimal_names)



import re
import tempfile

import unittest
from pathlib import Path

from click.testing import CliRunner

from ..cli import main

MOD_RE = re.compile(r"^.*\/mod", re.M)


class CliTest(unittest.TestCase):
    def test_basic(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d).resolve()
            (pd / "mod").mkdir()
            (pd / "mod" / "foo.py").write_text(
                """\
import sys
import click
import black
import bar
from baz import fromimport
"""
            )

            runner = CliRunner()
            result = runner.invoke(main, ["--requirements=requirements.txt", d])
            output = MOD_RE.sub("[TEMPDIR]/mod", result.output)
            self.assertEqual(
                """\
[TEMPDIR]/mod/foo.py uses bar but there is nothing installed to provide it
[TEMPDIR]/mod/foo.py uses baz.fromimport but there is nothing installed to provide it
[TEMPDIR]/mod/foo.py uses black but 'black' not in requirements
""",
                output,
            )

            runner = CliRunner()
            result = runner.invoke(
                main, ["--requirements=requirements.txt", "--missing-projects-only", d]
            )
            output = MOD_RE.sub("[TEMPDIR]/mod", result.output)
            self.assertEqual(
                """\
[TEMPDIR]/mod/foo.py uses bar but there is nothing installed to provide it
[TEMPDIR]/mod/foo.py uses baz.fromimport but there is nothing installed to provide it
['black']
""",
                output,
            )

            runner = CliRunner()
            result = runner.invoke(
                main, ["--requirements=requirements.txt", "--verbose", d]
            )
            output = MOD_RE.sub("[TEMPDIR]/mod", result.output)
            self.assertEqual(
                """\
[TEMPDIR]/mod/foo.py:
[TEMPDIR]/mod/foo.py uses bar but there is nothing installed to provide it
[TEMPDIR]/mod/foo.py uses baz.fromimport but there is nothing installed to provide it
[TEMPDIR]/mod/foo.py uses black but 'black' not in requirements
  black available from ['black']
  click available from ['click']
  sys stdlib
""",
                output,
            )



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Set


@dataclass
class Dist:
    provided_names: Set[str]
    namespace_names: Set[str]

    @property
    def minimal_names(self) -> Set[str]:
        # This heuristic is based on looking at real-world examples, in
        # particular the google-api-* and google-cloud-* projects.
        done: Set[str] = set()
        for p in sorted(self.provided_names):
            for x in reversed(list(iterparents(p))):
                if x not in self.namespace_names and not any(
                    n.startswith(f"{x}.") for n in self.namespace_names
                ):
                    done.add(x)
                    break
            else:
                if (
                    p not in self.namespace_names
                    and not any(n.startswith(f"{p}.") for n in self.namespace_names)
                    and p not in done
                ):
                    done.add(p)
        return done


def iterparents(f: str) -> Generator[str, None, None]:
    while "." in f:
        f = f.rsplit(".", 1)[0]
        yield f


def analyze(distinfo_dir: Path) -> Dist:
    packages: Set[str] = set()
    namespace_packages: Set[str] = set()

    # This is in two phases because otherwise we'd need to set __init__.py
    # before all other .py entries under a dir
    for line in (distinfo_dir / "RECORD").read_text().splitlines(True):
        filename, _ = line.split(",", 1)
        if ".." in filename:
            # docutils 0.18.1 seems to include a bin dir this way
            continue

        # TODO pyc-only dists
        # TODO .so, .dll, maybe even .dylib
        # TODO platform tag removal
        if filename.endswith(".py"):
            # TODO PEP 376 says this is full-blown CSV
            package = filename[:-3].replace("/", ".")  # TODO is this right?
            if package.endswith(".__init__"):
                package = package[: -len(".__init__")]
            packages.add(package)

    for package in sorted(packages):
        # Dragons: this is intended to detect new-style namespaces that are
        # _subdirs_ not individual files.  This isn't true for
        # googleapis-common-protos 1.56.2 (and is arguably a bug) but
        # consider this "best effort" code without perfect knowledge.
        for p in iterparents(package):
            if p not in packages:
                # print(f"add namespace for {p} {packages}")
                namespace_packages.add(p)

    namespace_package_file = distinfo_dir / "namespace_packages.txt"
    if namespace_package_file.exists():
        text = namespace_package_file.read_text()
        # pycodestyle==2.8.0 has just newlines in this file
        for line in text.splitlines():
            line = line.strip()
            if line:
                namespace_packages.add(line)

    return Dist(packages - namespace_packages, namespace_packages)


def main(dirname: str) -> None:  # pragma: no cover
    dist = analyze(Path(dirname))
    print("provided names=")
    for p in sorted(dist.provided_names):
        print(f"  {p}")
    print("namespace_names=")
    for p in sorted(dist.namespace_names):
        print(f"  {p}")
    print("minimal_names=")
    for p in sorted(dist.minimal_names):
        print(f"  {p}")


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv[1])



from .cli import CliTest
from .distinfo import IterDistinfoDirsTest
from .distinfo_inference import DistinfoInferenceTest
from .import_parser import ImportParserTest

__all__ = [
    "DistinfoInferenceTest",
    "IterDistinfoDirsTest",
    "ImportParserTest",
    "CliTest",
]



import ast
from pathlib import Path
from typing import Set


def get_imports(path: Path) -> Set[str]:
    tree = ast.parse(path.read_bytes())  # TODO can we get away with this
    imports: Set[str] = set()

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for subnode in node.names:
                imports.add(subnode.name)
        elif isinstance(node, ast.ImportFrom):
            if node.level == 0:
                for subnode in node.names:
                    imports.add(f"{node.module}.{subnode.name}")
    return imports



import tempfile
import unittest
from pathlib import Path

from ..import_parser import get_imports


class ImportParserTest(unittest.TestCase):
    def test_imports(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            pd = Path(d)
            (pd / "foo.py").write_text(
                """\
import a
import a.b
from a.b.c import z
from . import x  # relative imports are not parsed
from .x import y
"""
            )
            self.assertEqual({"a", "a.b", "a.b.c.z"}, get_imports(pd / "foo.py"))



# There are multiple projects that can answer the same questions this module
# does, but they are generally complex and have complex deps.
#
# For the goal of running this against a venv on CI, we don't need any of that
# and can rely on files being materialized on disk.

import re
import sys
from pathlib import Path
from typing import Generator, Tuple

from packaging.utils import canonicalize_name

DISTINFO_RE = re.compile(r"([^-]+)-(.*?)\.dist-info$")


def iter_all_distinfo_dirs() -> Generator[Tuple[str, str, Path], None, None]:
    for p in sys.path:
        path = Path(p)
        if path.is_dir():
            yield from iter_distinfo_dirs(path)


def iter_distinfo_dirs(path: Path) -> Generator[Tuple[str, str, Path], None, None]:
    for subdir in path.iterdir():
        if subdir.name.endswith(".dist-info") and subdir.is_dir():
            m = DISTINFO_RE.match(subdir.name)
            if not m:  # pragma: no cover
                continue
            (project, version) = m.groups()

            # Change from underscores to dashes
            project = canonicalize_name(project)

            yield project, version, subdir


if __name__ == "__main__":  # pragma: no cover
    for x in iter_all_distinfo_dirs():
        print(x)



import sys
from pathlib import Path
from typing import Dict, List, Optional, Set

import click

import trailrunner
from stdlibs import stdlib_module_names

from .distinfo import iter_all_distinfo_dirs, iter_distinfo_dirs

from .distinfo_inference import analyze, iterparents

from .import_parser import get_imports
from .requirements import iter_glob_all_requirement_names

STDLIB_MODULE_NAMES = stdlib_module_names()  # for the running version only


@click.command()
@click.option(
    "--requirements",
    default="requirements*.txt",
    help="Patterns for finding files from which to read requirements (comma-separated)",
    show_default=True,
)
@click.option("--verbose", "-v", is_flag=True, help="Show more logging")
@click.option("--installed-path", help="Where to look for distinfo if not sys.path")
@click.option(
    "--allow-names",
    help="Minimal names to consider satisfied, such as the current project top-level name (comma-separated)",
)
@click.option(
    "--missing-projects-only", is_flag=True, help="Show names of missing projects only"
)
@click.argument("target_dir")
def main(
    requirements: str,
    target_dir: str,
    installed_path: str,
    allow_names: Optional[str],
    verbose: bool,
    missing_projects_only: bool,
) -> None:
    available_names: Dict[str, Optional[str]] = {}
    requirement_names: Set[Optional[str]] = {None}

    # Part 1
    requirement_names = set(iter_glob_all_requirement_names(requirements))

    # Part 2
    if not installed_path:
        for (p, v, d) in iter_all_distinfo_dirs():
            dist = analyze(d)
            for name in dist.minimal_names:
                available_names[name] = p
    else:  # pragma: no cover
        for (p, v, d) in iter_distinfo_dirs(Path(installed_path)):
            dist = analyze(d)
            for name in dist.minimal_names:
                available_names[name] = p

    # Part 2b (first-party names, even if they're installed)
    if allow_names:
        for name in allow_names.split(","):
            available_names[name] = None

    # Part 3
    missing_projects: Set[str] = set()
    for path in trailrunner.walk(Path(target_dir)):
        if verbose:
            print(f"{path.as_posix()}:")
        imports = get_imports(path)
        for i in sorted(imports):
            if i in available_names or any(
                x in available_names for x in iterparents(i)
            ):
                providers: List[Optional[str]] = []
                for name in (i,) + tuple(iterparents(i)):
                    if name in available_names:
                        providers.append(available_names[name])

                if None in providers:  # first-party probably
                    continue

                if not any(p in requirement_names for p in providers):
                    # mypy doesn't know we've already checked for None above
                    missing_projects.update(providers)  # type: ignore
                    if not missing_projects_only:
                        if len(providers) == 1:
                            click.echo(
                                f"{path.as_posix()} uses "
                                + click.style(i, bold=True)
                                + " but "
                                + click.style(repr(providers[0]), bold=True)
                                + " not in requirements"
                            )
                        else:
                            click.echo(
                                f"{path.as_posix()} uses "
                                + click.style(i, bold=True)
                                + " but "
                                + click.style(repr(providers), bold=True)
                                + " not all in requirements"
                            )
                if verbose:
                    print(f"  {i} available from {providers}")
            elif i.split(".")[0] in STDLIB_MODULE_NAMES:
                if verbose:
                    print(f"  {i} stdlib")
            else:
                # TODO this might go to stderr, especially for
                # missing-projects-only mode
                click.echo(
                    f"{path.as_posix()} uses "
                    + click.style(i, bold=True)
                    + " but there is nothing installed to provide it"
                )
    if missing_projects_only:
        print(sorted(missing_projects))
    if missing_projects:
        sys.exit(1)


if __name__ == "__main__":
    main()
